# Copyright 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import os
import unittest

from testing import testcommon
import gvn.util
import gvn.errors


class TestGvnUtil(unittest.TestCase):

  SOME_BAD_CHANGE_NAME_CHARS = '#?&;,+=*%$!()'

  def testParseChangeName(self):
    self.assertEquals(('sybil', 'cool_feature', 1234),
                      gvn.util.ParseChangeName('sybil/cool_feature@1234'))
    self.assertEquals((None, 'cool_feature', None),
                      gvn.util.ParseChangeName('cool_feature@q'))
    self.assertEquals((None, 'cool_feature', -1),
                      gvn.util.ParseChangeName('cool_feature@-1'))
    self.assertEquals((None, 'cool_feature', 1234),
                      gvn.util.ParseChangeName('/cool_feature@1234'))
    self.assertEquals((None, 'cool_feature', 1234),
                      gvn.util.ParseChangeName('cool_feature@1234'))
    self.assertEquals((None, 'cool_feature', 1234),
                      gvn.util.ParseChangeName('cool_feature@r1234'))
    self.assertEquals(('sybil', 'cool_feature', None),
                      gvn.util.ParseChangeName('sybil/cool_feature@'))
    self.assertEquals(('sybil', 'cool_feature', None),
                      gvn.util.ParseChangeName('sybil/cool_feature'))
    self.assertEquals((None, 'cool_feature', None),
                      gvn.util.ParseChangeName('/cool_feature'))
    self.assertEquals((None, 'cool_feature', None),
                      gvn.util.ParseChangeName('/cool_feature@'))
    self.assertEquals((None, 'cool_feature', None),
                      gvn.util.ParseChangeName('cool_feature@'))
    self.assertEquals((None, 'cool_feature', None),
                      gvn.util.ParseChangeName('cool_feature'))
    self.assertEquals((None, None, None),
                      gvn.util.ParseChangeName(''))
    self.assertEquals((None, None, None),
                      gvn.util.ParseChangeName(None))

  def testValidateChangeName(self):
    # Test that InvalidChangeName gets raised appropriately.
    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, None)
    self.assert_(not gvn.util.IsValidChangeName(None))

    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, '')
    self.assert_(not gvn.util.IsValidChangeName(''))

    for c in self.SOME_BAD_CHANGE_NAME_CHARS:
      path = "bad%sname" % c
      self.assertRaises(gvn.errors.InvalidChangeName,
                        gvn.util.ValidateChangeName, path)
      self.assert_(not gvn.util.IsValidChangeName(path))

    # Test change names that might look like repository revisions.
    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, '1234')
    self.assertFalse(gvn.util.IsValidChangeName('1234'))
    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, 'r1234')
    self.assert_(not gvn.util.IsValidChangeName('r1234'))
    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, 'R1234')
    self.assert_(not gvn.util.IsValidChangeName('R1234'))

    long_change_name_50 = "a1234567890123456789012345678901234567890123456789"
    gvn.util.ValidateChangeName(long_change_name_50)
    self.assert_(gvn.util.IsValidChangeName(long_change_name_50))

    long_change_name_51 = "a12345678901234567890123456789012345678901234567890"
    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, long_change_name_51)
    self.assert_(not gvn.util.IsValidChangeName(long_change_name_51))

    # Test some change names that can be mistaken for existing
    # files or directories.
    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, '.')
    self.assert_(not gvn.util.IsValidChangeName('.'))

    self.assertRaises(gvn.errors.InvalidChangeName,
                      gvn.util.ValidateChangeName, '..')
    self.assert_(not gvn.util.IsValidChangeName('..'))

  def testMatchRevision(self):
    self.assertEquals('1234', gvn.util.MatchRevision('1234'))
    self.assertEquals('1234', gvn.util.MatchRevision('r1234'))
    self.assertEquals('1234', gvn.util.MatchRevision('R1234'))

    text = 'some test r5678'
    self.assertEquals(None, gvn.util.MatchRevision(text))
    self.assertEquals(None, gvn.util.MatchRevision(text[9:]))
    self.assertEquals('5678', gvn.util.MatchRevision(text[10:]))
    self.assertEquals('5678', gvn.util.MatchRevision(text[11:]))
    text = gvn.util.MatchRevision(text[10:])
    self.assertEquals('5678', text)

    self.assertEquals(None, gvn.util.MatchRevision('x1234'))
    self.assertEquals(None, gvn.util.MatchRevision(None))
    self.assertEquals(None, gvn.util.MatchRevision(''))

  def testRelativePath(self):
    self.assertRaises(gvn.errors.PathNotChild,
                      gvn.util.RelativePath, 'tmp', 'foo')
    self.assertRaises(gvn.errors.PathNotChild,
                      gvn.util.RelativePath, '/tmp', 'foo')
    self.assertRaises(gvn.errors.PathNotChild,
                      gvn.util.RelativePath, '/tmp', '/etc')
    self.assertEquals(gvn.util.RelativePath('/tmp/wc', '/tmp/wc/lib'), 'lib')
    self.assertEquals(gvn.util.RelativePath('tmp/wc', 'tmp/wc/lib'), 'lib')
    self.assertEquals(gvn.util.RelativePath('/tmp/wc', '/tmp/wc/tmp/wclib'),
                      'tmp/wclib')
    self.assertEquals(gvn.util.RelativePath('/tmp/wc', '/tmp/wc/tmp/wclib'),
                      'tmp/wclib')
    self.assertEquals(gvn.util.RelativePath('/tmp/wc', '/tmp/wc/lib/tmp/wc'),
                      'lib/tmp/wc')
    self.assertEquals(gvn.util.RelativePath('', 'foo'), 'foo')
    self.assertEquals(gvn.util.RelativePath('', 'foo/bar'), 'foo/bar')
    self.assertRaises(gvn.errors.PathNotChild,
                      gvn.util.RelativePath, '', '/etc')

  def testIsChild(self):
    self.assertEquals(gvn.util.IsChild('foo', 'foo'), True)
    self.assertEquals(gvn.util.IsChild('foo', 'bar'), False)
    self.assertEquals(gvn.util.IsChild('bar/foo', 'bar'), True)
    self.assertEquals(gvn.util.IsChild('foo/bar', 'bar'), False)
    self.assertEquals(gvn.util.IsChild('foobar', 'foo'), False)
    self.assertEquals(gvn.util.IsChild('foo', ''), True)
    self.assertEquals(gvn.util.IsChild('/foo', ''), False)

  def testPathSplit(self):
    self.assertEquals(gvn.util.PathSplit(''), [''])
    self.assertEquals(gvn.util.PathSplit('/'), ['', ''])
    self.assertEquals(gvn.util.PathSplit('/tmp'), ['', 'tmp'])
    self.assertEquals(gvn.util.PathSplit('tmp/a/b'), ['tmp', 'a', 'b'])
    self.assertEquals(gvn.util.PathSplit('tmp/a/b', maxsplit=1),
                      ['tmp', 'a/b'])
    self.assertEquals(gvn.util.PathSplit('tmp/a/b', maxsplit=2),
                      ['tmp', 'a', 'b'])
    self.assertEquals(gvn.util.PathSplit('/tmp/a/b', maxsplit=1),
                      ['', 'tmp/a/b'])
    self.assertEquals(gvn.util.PathSplit('/tmp/a/b', maxsplit=2),
                      ['', 'tmp', 'a/b'])

  def testCommonPrefix(self):
    self.assertEquals(gvn.util.CommonPrefix([]), '')
    self.assertEquals(gvn.util.CommonPrefix(['', '']), '')
    self.assertEquals(gvn.util.CommonPrefix(['a', 'b']), '')
    self.assertEquals(gvn.util.CommonPrefix(['', 'tmp']), '')
    self.assertEquals(gvn.util.CommonPrefix(['/', '/tmp']), '/')
    self.assertEquals(gvn.util.CommonPrefix(['/', '/']), '/')
    self.assertEquals(gvn.util.CommonPrefix(['/tmp', '/tmp']), '/tmp')
    self.assertEquals(gvn.util.CommonPrefix(['/tmp', '/tmp/a']), '/tmp')
    self.assertEquals(gvn.util.CommonPrefix(['/a/b/c/d', '/a/b/c/e/f']),
                      '/a/b/c')
    self.assertEquals(gvn.util.CommonPrefix(['/a/b/c/d', '/a/b/c',
                                             '/a/b/c/e/f']), '/a/b/c')
    self.assertEquals(gvn.util.CommonPrefix(['/tmp/foo', '/tmp/bar']), '/tmp')
    self.assertEquals(gvn.util.CommonPrefix(['/tmp/ab', '/tmp/ac']), '/tmp')
    self.assertEquals(gvn.util.CommonPrefix(['/tmp/a', '/tmp/ac']), '/tmp')
    self.assertEquals(gvn.util.CommonPrefix(['/tmp/a', '/tmp/ab', '/tmp/ac']),
                      '/tmp')
    self.assertEquals(gvn.util.CommonPrefix(['/A/D/G/rho',
                                             '/A']), '/A')


  def testConvertStringDateToDateTime(self):
    convert = gvn.util.ConvertStringDateToDateTime
    self.assertEquals(convert('2007-04-19T20:15:28.654321Z'),
                      datetime.datetime(2007, 4, 19, 20, 15, 28))
    self.assertEquals(convert('1995-12-13T03:04:05.0Z'),
                      datetime.datetime(1995, 12, 13, 3, 4, 5))

  def testClimbAndFind(self):
    def find_root(path):
      return -1
    def find_foo(path):
      if path.endswith('foo'):
        return 0
      return -1
    def find_child_of_foo(path):
      if path.endswith('foo'):
        return 1
      return -1
    self.assertEquals(gvn.util.ClimbAndFind('/foo/bar/baz', find_root), '/')
    self.assertEquals(gvn.util.ClimbAndFind('foo/bar/baz', find_root), '/')
    self.assertEquals(gvn.util.ClimbAndFind('/foo/bar/baz', find_foo),
                      '/foo')
    self.assertEquals(gvn.util.ClimbAndFind('/foo/bar/baz', find_child_of_foo),
                      '/foo/bar')
    self.assertEquals(gvn.util.ClimbAndFind('foo/bar/baz', find_root), '/')
    self.assertEquals(gvn.util.ClimbAndFind('C:\\foo\\bar', find_root), '/')

  def testEditor(self):
    editor_name = tmp_prefix = 'testEditor'
    def get_tmpfile(command):
      (empty, tmpfile) = command.split(editor_name + ' ')
      self.assertEqual(empty, '',
                       msg="command not '%s' in '%s'" % (editor_name, command))
      return tmpfile

    editor = gvn.util.Editor(editor_name)
    self.assertTrue(editor.IsDone())

    # basic no-op edit
    tmpfile = []
    def edit(command):
      tmpfile.append(get_tmpfile(command))
      return 0
    self.assertEqual(editor.Edit('some text', 'utf8', tmp_prefix, edit),
                     'some text')
    self.assertEqual(tmpfile, [editor.tmpfile])
    self.assertEqual(open(editor.tmpfile).read(), 'some text')
    editor.Done()
    self.assertEqual(editor.tmpfile, None)
    self.assertFalse(os.path.exists(tmpfile[0]))

    # edited file
    tmpfile = []
    def edit(command):
      tmpfile.append(get_tmpfile(command))
      fp = open(tmpfile[0], 'w')
      fp.write('some other text')
      fp.close()
      return 0
    self.assertEqual(editor.Edit('some text', 'utf8', tmp_prefix, edit),
                     'some other text')
    self.assertEqual(tmpfile, [editor.tmpfile])
    self.assertEqual(open(editor.tmpfile).read(), 'some other text')
    editor.Done()
    self.assertEqual(editor.tmpfile, None)
    self.assertFalse(os.path.exists(tmpfile[0]))

    # non-zero editor exit status
    tmpfile = []
    def edit(command):
      tmpfile.append(get_tmpfile(command))
      return 2
    self.assertRaises(gvn.errors.Editor,
                      editor.Edit, 'some text', 'utf8', tmp_prefix, edit)
    self.assertEqual(tmpfile, [editor.tmpfile])
    self.assertEqual(open(editor.tmpfile).read(), 'some text')
    editor.Done()
    self.assertEqual(editor.tmpfile, None)
    self.assertFalse(os.path.exists(tmpfile[0]))

    # truncated file, indicating user abort
    tmpfile = []
    def edit(command):
      tmpfile.append(get_tmpfile(command))
      fp = open(tmpfile[0], 'w')
      fp.close()
      return 0
    self.assertEqual(editor.Edit('some text', 'utf8', tmp_prefix, edit),
                     None)
    self.assertEqual(tmpfile, [editor.tmpfile])
    self.assertEqual(open(editor.tmpfile).read(), '')
    editor.Done()
    self.assertEqual(editor.tmpfile, None)
    self.assertFalse(os.path.exists(tmpfile[0]))

    # removed file, indicating user abort
    tmpfile = []
    def edit(command):
      tmpfile.append(get_tmpfile(command))
      os.unlink(tmpfile[0])
      return 0
    self.assertEqual(editor.Edit('some text', 'utf8', tmp_prefix, edit),
                     None)
    self.assertEqual(None, editor.tmpfile)
    self.assertFalse(os.path.exists(tmpfile[0]))
    editor.Done()
    self.assertEqual(editor.tmpfile, None)
    self.assertFalse(os.path.exists(tmpfile[0]))
